/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2003, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

import java.util.*;

/**
 * A reference to another attribute.
 * A leaf node in an expression tree that represents an attribute reference.
 * @see Expr
 * @see RecordExpr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 2.0
 */
public class AttrRef extends Expr {
    private static String VERSION = "$Id: AttrRef.java,v 1.12 2003/01/20 21:05:03 solomon Exp $";

    /** The string name associated with this reference. */
    public final AttrName name;

    /** Create a new AttrRef.
     * @param name the string name associated with this reference.
     */
    public AttrRef(AttrName name) {
        super(ATTRIBUTE);
        this.name = name;
    } // AttrRef(AttrName)

    /** Create a new AttrRef.
     * @param name the string name associated with this reference.
     */
    public AttrRef(String name) {
        this(AttrName.fromString(name));
    } // AttrRef(AttrName)

    /** Convert this AttrRef to a string.
     */
    public String toString() {
        return name.toString();
    } // toString()

    /** Convert this Expr to a string, appending the result to the end of "sb".
     * The representation is the "canonical native format".
     * @param sb a place to put the result.
     * @return sb.
     * @see ClassAdWriter
     */
    public StringBuffer toString(StringBuffer sb) {
        return sb.append(name);
    } // toString(StringBuffer)

    /** Evaluate this Expr.  This is the internal method used to implement
     * {@link Expr#eval()}.
     * The env is searched for name of this AttrRef.  Scopes that do not
     * define that name are popped off the stack, leaving the scope defining
     * the name at the top.  If the name is not found in any of the scopes,
     * the env becomes the empty environment, and the result is the UNDEFINED
     * constant.
     * @param env the environment to search for the name.  It is modified in
     * place to indicate where the name was resolved.
     * @return a Constant, AttrRef, ListExpr, or RecordExpr representing the
     * value.
     */
    protected Expr eval1(Env env) {
        Expr target = env.search(name);
        if (target == null) {
            return Constant.undefined("attribute " + name + " not found");
        } else {
            return target;
        }
    } // eval1(Env)

    /** The precedence of this expression
     * @return the precendence of this node (MAXPREC).
     */
    public int prec() {
        return MAXPREC;
    } // prec()
} // AttrRef
