/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2003, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

import java.util.*;
import java.io.*;

/**
 * A library of handy methods for manipulating classad expressions.
 * The class contains only static methods for performing various operations on
 * classified advertisements.  See {@link Expr} and its various subclasses,
 * especially {@link RecordExpr} for the actual objects that are manipulated.
 * @see Expr
 * @see RecordExpr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 2.0
 */
public class ClassAd {
    private static String VERSION = "$Id: ClassAd.java,v 1.8 2003/01/20 21:05:03 solomon Exp $";

    // Convenience functions for matching class ads.

    /** The default constructor is made private to prevent anybody from trying
     * to create an instance of this class.
     */
    private ClassAd() {
        throw new RuntimeException();
    } // ClassAd()

    /** Wrap a pair of ads in an environment and context such that inside
     * ad1, <code>self</code> refers to <code>ad1</code>, <code>other</code>
     * refers to <code>ad2</code>, and any other attribute is searched for
     * first in <code>ad1</code> and then in <code>ad2</code> (and similarly,
     * with <code>ad1</code> and <code>ad2</code> interchanged).
     * @param ad1 the first ClassAd.
     * @param ad2 the second ClassAd.
     * @return a <code>RecordExpr E</code> such that
     * <code>E.Ad1.self</code> is <code>ad1</code>, and
     * <code>E.Ad2.self</code> is <code>ad2</code>.
     */
     public static RecordExpr bind(RecordExpr ad1, RecordExpr ad2) {
        // Create the classad env =
        // [
        //    Ad1 = [other = Ad2.self; self = <ad1>]
        //    Ad2 = [other = Ad1.self; self = <ad2>]
        // ]
        RecordExpr env = new RecordExpr();
        RecordExpr env1 = new RecordExpr();
        env1.insertAttribute("other", (new AttrRef("Ad2")).selectExpr("self"))
            .insertAttribute("self", ad1);
        RecordExpr env2 = new RecordExpr();
        env2.insertAttribute("other", (new AttrRef("Ad1")).selectExpr("self"))
            .insertAttribute("self", ad2);
        env.insertAttribute("Ad1", env1);
        env.insertAttribute("Ad2", env2);

        return env;
    } // bind(RecordExpr,RecordExpr)

    /** Evaluate a selection from a ClassAd.
     * @param ad a ClassAd from which to select.
     * @param attrs a sequence <code>attr1, attr2, ..., attrn</code> of
     * attribute names.
     * @return the result of evaluating <code>ad.attr1.attr2...attrn</code>.
     */
    public static Expr eval(RecordExpr ad, String[] attrs) {
        Expr sel = ad;
        for (int i=0; i<attrs.length; i++) {
            sel = sel.selectExpr(attrs[i]);
        }
        return sel.eval();
    } // eval(RecordExpr,String[])

    /** Evaluate a selection from a ClassAd.
     * @param ad a ClassAd from which to select.
     * @param attr an attribute name.
     * @return the result of evaluating <code>ad.attr</code>.
     */
    public static Expr eval(RecordExpr ad, String attr) {
        Expr sel = ad.selectExpr(attr);
        return sel.eval();
    } // eval(RecordExpr,String[])

    /** Evaluate an expression in the context of a ClassAd.
     * The <code>expr</code> is evaluated as if it were a the value of an
     * attribute named <code>name</code> in <code>ad</code>.
     * <p>
     * <b>Note</b>: the RecordExpr <code>ad</code> is modified and then
     * returned to its original state, so this method is not thread-safe.
     * @param expr the expression to evaluate.
     * @param ad the context for evaluation.
     * @return the result of evaluating <code>expr</code> in the context of
     *         <code>ad</code>.
     */
    public static Expr eval(String name, Expr expr, RecordExpr ad) {
        AttrName n = AttrName.fromString(name);
        ad.insertAttribute(n, expr);
        Expr result = ad.selectExpr(name).eval();
        ad.removeAttribute(n);
        return result;
    } // eval(Expr,RecordExpr)

    /** Evaluate an attribute of a ClassAd in an environment consisting of
     * a pair of ads. The call <code>ClassAd.eval(attr, ad1, ad2)</code>
     * evaluates <code>ad1.attr</code> an environment such that inside
     * ad1, <code>self</code> refers to <code>ad1</code>, <code>other</code>
     * refers to <code>ad2</code>, and any other attribute is searched for
     * first in <code>ad1</code> and then in <code>ad2</code> (and similarly,
     * with <code>ad1</code> and <code>ad2</code> interchanged).
     * @param attr the name of the attribute used to initiate the evaluation
     * @param ad1 the ClassAd used to initiate the evaluation
     * @param ad2 the other ClassAd
     * @return the result of the evaluation
     */
    public static Expr eval(String attr, RecordExpr ad1, RecordExpr ad2) {
        return eval(bind(ad1, ad2), new String[] { "Ad1", "self", attr });
    } // eval(String,RecordExpr,RecordExpr)

    /** Match two ClassAds.  If each expr is a RecordExpr with a
     * <code>Requirements</code> attribute that evaluates to <b>true</b> in the
     * context of the other expr and each expr has a <code>Rank</code>
     * attribute that evaluates to an integer, the result is two-element array
     * containing the <code>Rank</code> values.  In all other cases, the result
     * is <b>null</b>.
     * @param expr1 the first expression.
     * @param expr2 the second expression.
     * @return the result of the match, either a pair of integers or
     * <b>null</b>.
     */
    public static int[] match(Expr expr1, Expr expr2) {
        if (expr1.type != Expr.RECORD || expr2.type != Expr.RECORD) {
            return null;
        }

        // Create an evaluation environment
        RecordExpr ad = bind((RecordExpr) expr1, (RecordExpr) expr2);

        // Check requirements
        if (!eval(ad, AD1_SELF_REQUIREMENTS).isTrue()
            || !eval(ad, AD2_SELF_REQUIREMENTS).isTrue())
        {
            return null;
        }

        // Evaluate ranks
        try {
            return new int[] {
                eval(ad, AD1_SELF_RANK).intValue(),
                eval(ad, AD2_SELF_RANK).intValue()
            };
        } catch (ArithmeticException e) {
            return null;
        }
    } // match(Expr,Expr)

    /** Path expression for use in match(). */
    private static final String[] AD1_SELF_REQUIREMENTS =
                                new String[] { "Ad1", "self", "requirements" };
    /** Path expression for use in match(). */
    private static final String[] AD2_SELF_REQUIREMENTS =
                                new String[] { "Ad2", "self", "requirements" };
    /** Path expression for use in match(). */
    private static final String[] AD1_SELF_RANK =
                                new String[] { "Ad1", "self", "rank" };
    /** Path expression for use in match(). */
    private static final String[] AD2_SELF_RANK =
                                new String[] { "Ad2", "self", "rank" };

    // Factories (static methods returning new objects)
    // See also ClassAdParser

    /** Create a constant expression from an integer value.
     * @param i the value.
     * @return the resulting Constant object.
     */
    public static Constant constant(int i) {
        return Constant.getInstance(i);
    } // constant(int)

    /** Create a constant expression from a real value.
     * @param x the value.
     * @return the resulting Constant object.
     */
    public static Constant constant(double x) {
        return Constant.getInstance(x);
    } // constant(double)

    /** Create a constant expression from a String value.
     * @param s the value.
     * @return the resulting Constant object.
     */
    public static Constant constant(String s) {
        return Constant.getInstance(s);
    } // constant(String)

} // ClassAd
