/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2003, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

import java.io.*;
import java.util.Date;
import java.util.Iterator;
import java.util.SimpleTimeZone;
import java.text.DateFormat;

/** A tool for converting classads to characters in a variety of formats.
 * A ClassAdWriter behaves like a java.io.PrintWriter except that it has two
 * additional methods <code>print(Expr)</code> and <code>println(Expr)</code>.
 * An instance has an associated <em>representation</em> and set of
 * <em>options</em>.
 * <p>
 * Currently, we support three representations:
 * <dl>
 * <dt>NATIVE<dd>The "native" ASCII representation, e.g.
 * <pre>
 *    [ my_string = "abc"; my_list = {1, 2, x + 1};]
 * </pre>
 *
 * <dt>XML<dd>An XML (Extensible Markup Language) representation, e.g.
 * <pre>
 *     &lt;c&gt;
 *         &lt;a n="my_string"&gt;&lt;s&gt;abc&lt;/s&gt;&lt;/a&gt;
 *         &lt;a n="my_list"&gt;&lt;l&gt; &lt;n&gt;1&lt;/n&gt; &lt;n&gt;2&lt;/n&gt; &lt;e&gt;x+1&lt;/e&gt; &lt;/l&gt;&lt;/a&gt;
 *     &lt;/c&gt;
 * </pre>
 * </dl>
 *
 * The third representation is "OLD" classads, supported only for compatibility
 * with a previous version C++ implementation of classified advertisements.
 * Only RecordExprs can be represented in this format, and the format is
 * inherently binary, requiring a DataOutput.  To represent an ad in the OLD
 * format, use {@link condor.classad.RecordExpr#transmit(java.io.DataOutput)}.
 * <p>
 * Other external representations may be added in the future.
 * <p>
 * Options provide finer control over the output format, for example, the
 * native representation may be formatted for readability (e.g. listing each
 * attribute of a RecordExpression on a separate line) or for compactness.
 * By default, a ClassAdWriter uses NATIVE representation and formats the same
 * as <code>Expr.toString()</code> so that code such as
 * <pre>
 *     Expr e = ...;
 *     ClassAdWriter out = new ClassAdWriter(System.out);
 *     out.println(e)
 * </pre>
 * has the same effect as
 * <pre>
 *     Expr e = ...;
 *     System.out.println(e);
 * </pre>
 * However, a ClassAdWriter allows more control over representation and
 * formatting details.  It also may be more efficient for large expressions,
 * since the string representation doesn't need to be assembled in memory
 * before being printed.
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 2.0
 */
public class ClassAdWriter extends java.io.PrintWriter {
    private static String VERSION = "$Id: ClassAdWriter.java,v 1.3 2003/01/20 21:05:03 solomon Exp $";

    // Constructors

    /** Creates a new ClassAdWriter for writing to an existing OutputStream
     * in a given representation.
     * Characters are converted to bytes using the default character encoding.
     * @param out the desitination for output.
     * @param representation the representation (one of NATIVE or XML).
     * @param autoFlush if true, the println() methods will flush the output
     * buffer.
     */
    public ClassAdWriter(OutputStream out,
                         int representation,
                         boolean autoFlush)
    {
        super(out, autoFlush);
        rep = representation;
        if (rep == XML) {
            println("<?xml version=\"1.0\"?>");
            println("<!DOCTYPE classads>");
            println("<classads>");
        }
    } // ClassAdWriter(OutputStream,int,boolean)

    /** Creates a new ClassAdWriter for writing to an existing OutputStream
     * in a given representation, without automatic line flushing.
     * Characters are converted to bytes using the default character encoding.
     * @param out the desitination for output.
     * @param representation the representation (one of NATIVE or XML).
     */
    public ClassAdWriter(OutputStream out, int representation) {
        this(out, representation, false);
    } // ClassAdWriter(OutputStream,int)

    /** Creates a new ClassAdWriter for writing to an existing OutputStream
     * using the NATIVE representation.
     * Characters are converted to bytes using the default character encoding.
     * @param out the desitination for output.
     * @param autoFlush if true, the println() methods will flush the output
     * buffer.
     */
    public ClassAdWriter(OutputStream out, boolean autoFlush) {
        super(out, autoFlush);
        rep = NATIVE;
    } // ClassAdWriter(OutputStream,boolean)

    /** Creates a new ClassAdWriter for writing to an existing OutputStream
     * using the NATIVE representation, without automatic line flushing.
     * Characters are converted to bytes using the default character encoding.
     * @param out the desitination for output.
     */
    public ClassAdWriter(OutputStream out) {
        super(out);
        rep = NATIVE;
    } // ClassAdWriter(OutputStream)

    /** Creates a new ClassAdWriter for writing to an existing Writer
     * in a given representation.
     * Characters are converted to bytes using the default character encoding.
     * @param out the desitination for output.
     * @param representation the representation (one of NATIVE or XML).
     * @param autoFlush if true, the println() methods will flush the output
     * buffer.
     */
    public ClassAdWriter(Writer out,
                         int representation,
                         boolean autoFlush)
    {
        super(out, autoFlush);
        rep = representation;
        if (rep == XML) {
            println("<?xml version=\"1.0\"?>");
            println("<!DOCTYPE classads>");
            println("<classads>");
        }
    } // ClassAdWriter(Writer,int,boolean)

    /** Creates a new ClassAdWriter for writing to an existing Writer
     * in a given representation, without automatic line flushing.
     * Characters are converted to bytes using the default character encoding.
     * @param out the desitination for output.
     * @param representation the representation (one of NATIVE or XML).
     */
    public ClassAdWriter(Writer out, int representation) {
        this(out, representation, false);
    } // ClassAdWriter(Writer,int)

    /** Creates a new ClassAdWriter for writing to an existing Writer
     * using the NATIVE representation.
     * Characters are converted to bytes using the default character encoding.
     * @param out the desitination for output.
     * @param autoFlush if true, the println() methods will flush the output
     * buffer.
     */
    public ClassAdWriter(Writer out, boolean autoFlush) {
        super(out, autoFlush);
        rep = NATIVE;
    } // ClassAdWriter(Writer,boolean)

    /** Creates a new ClassAdWriter for writing to an existing Writer
     * using the NATIVE representation, without automatic line flushing.
     * Characters are converted to bytes using the default character encoding.
     * @param out the desitination for output.
     */
    public ClassAdWriter(Writer out) {
        super(out);
        rep = NATIVE;
    } // ClassAdWriter(Writer)

    // Public constants

    /** Native output representation (e.g. <code>[a=x+1]</code>). */
    public static final int NATIVE = 1;

    /** XML output representation
     * (e.g.  <code>&lt;c&gt;&lt;a name="a"&gt;&lt;e&gt;x+1&lt;/e&gt;&lt;/a&gt;&lt;/c&gt; </code>).
     */
    public static final int XML = 2;

    /** Flags to control the formatting of expressions. A value of zero creates
     * a representation that is unambiguous and easy to parse but not very
     * readable.  In the case of the NATIVE representation, this is the
     * "canonical" format.  Each bit turns on a formatting feature that makes
     * the string more readable, but perhaps more verbose and/or ambiguous.
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    private int formatFlags = 0;

    /** An option flag for <code>formatFlags</code>.
     * Don't surround strings with quotes or escape special characters in the
     * string.
     * <p><strong>WARNING</strong>: Strings created with this flag cannot be
     * parsed to recover the expressions that created them.
     * @see #setFormatFlags
     */
    static public final int NO_ESCAPE_STRINGS = 1<<0;

    /** An option flag for <code>formatFlags</code>.
     * Print the ERROR value as ``ERROR(reason)'' rather than just ``ERROR''.
     * @see #setFormatFlags
     */
    static public final int SHOW_ERROR_DETAIL = 1<<1;

    /** An option flag for <code>formatFlags</code>.
     * Use only as many parentheses are are required to override the default
     * precendence of operators.
     * @see #setFormatFlags
     */
    static public final int MINIMAL_PARENTHESES = 1<<2;

    /** An option flag for <code>formatFlags</code>.
     * Display classads with one attribute per line.
     * @see #setFormatFlags
     */
    static public final int MULTI_LINE_ADS = 1<<3;

    /** An option flag for <code>formatFlags</code>.
     * Display lists with one member per line.
     * @see #setFormatFlags
     */
    static public final int MULTI_LINE_LISTS = 1<<4;

    /** A combination of option flags for <code>formatFlags</code>.
     * Produce a concise version (both MINIMAL_PARENTHESES and
     * NO_ESCAPE_STRINGS).
     * @see #setFormatFlags
     * @see #MINIMAL_PARENTHESES
     * @see #NO_ESCAPE_STRINGS
     */
    static public final int BRIEF = MINIMAL_PARENTHESES | NO_ESCAPE_STRINGS;

    /** A combination of option flags for <code>formatFlags</code>.
     * Produce a compact but parsable version (currently, the same as
     * MINIMAL_PARENTHESES).
     * @see #setFormatFlags
     * @see #MINIMAL_PARENTHESES
     */
    static public final int COMPACT = MINIMAL_PARENTHESES;


    /** A combination of option flags for <code>formatFlags</code>.
     * Produce a maximally-readable version (all flags set).
     * @see #setFormatFlags
     */
    static public final int READABLE =
                                NO_ESCAPE_STRINGS
                                | SHOW_ERROR_DETAIL
                                | MINIMAL_PARENTHESES
                                | MULTI_LINE_ADS
                                | MULTI_LINE_LISTS;

    /** Set options for converting expressions to strings.
     * @param flag the new flag values.
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    public void setFormatFlags(int flags) {
        formatFlags = flags;
    } // setFormatFlags(int)

    /** Get existing options for converting expressions to strings.
     * @return the current option flags.
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    public int getFormatFlags() {
        return formatFlags;
    } // getFormatFlags()

    /** Turn on options for converting expressions to strings.
     * @param flag the flag values to be enabled
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    public void enableFormatFlags(int flags) {
        formatFlags |= flags;
    } // enableFormatFlags(int)

    /** Turn off options for converting expressions to strings.
     * @param flag the flag values to be disabled
     * @see #NO_ESCAPE_STRINGS
     * @see #SHOW_ERROR_DETAIL
     * @see #MINIMAL_PARENTHESES
     * @see #MULTI_LINE_ADS
     * @see #MULTI_LINE_LISTS
     * @see #BRIEF
     * @see #COMPACT
     * @see #READABLE
     */
    public void disableFormatFlags(int flags) {
        formatFlags &= ~flags;
    } // disableFormatFlags(int)


    // Internal state.

    /** The representation to be used for output. */
    private final int rep;

    /** The current indentation level, for "pretty printing". */
    private int indentLevel = 0;

    /** Indent output by the current indentation level. */
    private void indent() {
        for (int i = 0; i < indentLevel; i++) {
            print("    ");
        }
    } // indent

    // Public methods.

    /** Print a classad expression and then terminate the line. This method
     * behaves as though it invokes print(Expr) and then println().
     * @param exp the expression to be printed.
     */
    public void println(Expr exp) {
        print(exp);
        println();
    } // println(Expr)

    /** Print a classad expression.
     * @param exp the expression to be printed.
     */
    public void print(Expr exp) {
        switch (rep) {
        case NATIVE: printNative(exp, false); return;
        case XML: printXML(exp); return;
        default:
            throw new RuntimeException(
                "Internal error, unknown representation " + rep);
        }
    } // print(Expr)

    /** Close this writer. */
    public void close() {
        if (rep == XML) {
            println("</classads>");
        }
        super.close();
    } // close()

    /** Print a classad expression in NATIVE format.
     * @param exp the expression to be printed.
     * @param escapeXML if true, any occurrences of the four xml characters
     *      quote, less-than, greater-than or ampersand should be replaced by
     *      the corresponding XML "entities", because this call is in the
     *      context of a call of printXML (e.g., inside an &lt;e&gt; element).
     */
    private void printNative(Expr exp, boolean escapeXML) {
        if (exp == null) {
            super.print(exp);
            return;
        }
        Constant c = exp.isConstant() ? (Constant) exp : null;
        switch (exp.type) {

        // Internal nodes in tree

        case Expr.CALL: {
            FuncCall e = (FuncCall) exp;
            Expr[] args = e.args;
            print(e.func);
            print('(');
            for (int i = 0; i < args.length; i++) {
                if (i > 0) {
                    print(',');
                }
                printNative(args[i], escapeXML);
            }
            print(')');
            break;
        }
        case Expr.COND: {
            CondExpr e = (CondExpr) exp;
            if ((formatFlags & MINIMAL_PARENTHESES ) == 0) {
            } else {
                if (e.ec.prec() <= e.prec()) {
                    print('(');
                    printNative(e.ec, escapeXML);
                    print(')');
                } else {
                    printNative(e.ec, escapeXML);
                }
                print('?');
                printNative(e.et, escapeXML);
                print(':');
                printNative(e.ef, escapeXML);
            }
            break;
        }
        case Expr.LIST: {
            ListExpr e = (ListExpr) exp;
            if ((formatFlags & MULTI_LINE_LISTS ) == 0) {
                print('{');
                for (int i = 0; i < e.size(); i++) {
                    if (i > 0) {
                        print(',');
                    }
                    printNative(e.sub(i), escapeXML);
                }
                print('}');
            } else {
                println('{');
                indentLevel++;
                indent();
                for (int i = 0; i < e.size(); i++) {
                    if (i > 0) {
                        println(',');
                        indent();
                    }
                    printNative(e.sub(i), escapeXML);
                }
                println();
                indentLevel--;
                indent();
                print('}');
            }
            break;
        }
        case Expr.OP: {
            Op e = (Op) exp;
            String op = escapeXML ? Expr.opNameXML[e.op] : Expr.opName[e.op];
            if ((formatFlags & MINIMAL_PARENTHESES ) == 0) {
                // Fully parenthesize
                print('(');
                if (e.arg2 == null) {
                    // unary operator
                    print(op);
                }
                // left (or only) operand
                printNative(e.arg1, escapeXML);
                if (e.arg2 != null) {
                    // binary operator
                    print(op);

                    // right operand
                    printNative(e.arg2, escapeXML);
                }
                print(')');
            } else {
                // Only use as many parentheses as necessary
                if (e.arg2 == null) {
                    // unary operator
                    print(op);
                }

                // left (or only) operand
                if (e.arg1.prec() < e.prec()) {
                    print('(');
                    printNative(e.arg1, escapeXML);
                    print(')');
                } else {
                    printNative(e.arg1, escapeXML);
                }

                if (e.arg2 != null) {
                    // binary operator
                    print(op);

                    // right operand
                    if (e.arg2.prec() <= e.prec()) {
                        print('(');
                        printNative(e.arg2, escapeXML);
                        print(')');
                    } else {
                        printNative(e.arg2, escapeXML);
                    }
                }
            } // minimal parentheses
            break;
        }
        case Expr.RECORD: {
            RecordExpr e = (RecordExpr) exp;
            boolean first = true;

            if ((formatFlags & MULTI_LINE_ADS ) == 0) {
                print('[');
                for (Iterator i = e.attributes(); i.hasNext(); ) {
                    if (first) {
                        first = false;
                    } else {
                        print(';');
                    }
                    AttrName attr = (AttrName) i.next();
                    String s = attr.toString();
                    if (escapeXML) {
                        printXMLString(s);
                    } else {
                        print(s);
                    }
                    print("=");
                    printNative(e.lookup(attr), escapeXML);
                }
                print(']');
            } else {
                println('[');
                indentLevel++;
                indent();
                for (Iterator i = e.attributes(); i.hasNext(); ) {
                    if (first) {
                        first = false;
                    } else {
                        println(';');
                        indent();
                    }
                    AttrName attr = (AttrName) i.next();
                    String s = attr.toString();
                    if (escapeXML) {
                        printXMLString(s);
                    } else {
                        print(attr.toString());
                    }
                    print(" = ");
                    printNative(e.lookup(attr), escapeXML);
                }
                println();
                indentLevel--;
                indent();
                print(']');
            }
            break;
        }
        case Expr.SELECTION: {
            SelectExpr e = (SelectExpr) exp;
            printNative(e.base, escapeXML);
            print('.');
            print(e.selector);
            break;
        }
        case Expr.SUBSCRIPT: {
            SubscriptExpr e = (SubscriptExpr) exp;
            printNative(e.base, escapeXML);
            print('[');
            printNative(e.selector, escapeXML);
            print(']');
            break;
        }

        // Leaf nodes (constants)

        case Expr.ATTRIBUTE: {
            String s = exp.toString();
            if (escapeXML) {
                printXMLString(s);
            } else {
                print(s);
            }
            break;
        }
        case Expr.ABSOLUTE_TIME:
            print("absTime(\"");
            print(Expr.dateAndTimeFormat.format((Date) c.value));
            print("\")");
            break;
        case Expr.RELATIVE_TIME:
            print("relTime(\"");
            print(Constant.rel2str(((Long) c.value).longValue()));
            print("\")");
            break;
        case Expr.BOOLEAN:
            print(c == Constant.True ? "true" : "false");
            break;
        case Expr.ERROR:
            if ((formatFlags & SHOW_ERROR_DETAIL) == 0) {
                print("ERROR");
            } else {
                print("ERROR(");
                print(c.value);
                print(')');
            }
            break;
        case Expr.UNDEFINED:
            if ((formatFlags & SHOW_ERROR_DETAIL) == 0) {
                print("UNDEFINED");
            } else {
                print("UNDEFINED(");
                print(c.value);
                print(')');
            }
            break;
        case Expr.INTEGER:
        case Expr.REAL:
            print(c.value);
            break;
        case Expr.STRING:
            if ((formatFlags & NO_ESCAPE_STRINGS) == 0) {
                printQuotedString((String) c.value, '"', escapeXML);
            } else {
                print(c.value);
            }
            break;
        } // switch (exp.type)
    } // printNative(Expr)

    /** Print a classad expression in XML format.
     * @param exp the expression to be printed.
     */
    private void printXML(Expr exp) {
        if (exp == null) {
            super.print(exp);
            return;
        }
        Constant c = exp.isConstant() ? (Constant) exp : null;
        switch (exp.type) {
        case Expr.LIST: {
            ListExpr e = (ListExpr) exp;
            if ((formatFlags & MULTI_LINE_LISTS ) == 0) {
                print("<l>");
                for (int i = 0; i < e.size(); i++) {
                    printXML(e.sub(i));
                }
                print("</l>");
            } else {
                println("<l>");
                indentLevel++;
                for (int i = 0; i < e.size(); i++) {
                    indent();
                    printXML(e.sub(i));
                    println();
                }
                indentLevel--;
                indent();
                print("</l>");
            }
            break;
        }
        case Expr.RECORD: {
            RecordExpr e = (RecordExpr) exp;
            boolean first = true;

            if ((formatFlags & MULTI_LINE_ADS ) == 0) {
                print("<c>");
                for (Iterator i = e.attributes(); i.hasNext(); ) {
                    AttrName attr = (AttrName) i.next();
                    print("<a n=\"");
                    printXMLString(attr.rawString());
                    print("\">");
                    printXML(e.lookup(attr));
                    print("</a>");
                }
                print("</c>");
            } else {
                println("<c>");
                indentLevel++;
                for (Iterator i = e.attributes(); i.hasNext(); ) {
                    indent();
                    AttrName attr = (AttrName) i.next();
                    print("<a n=\"");
                    printXMLString(attr.rawString());
                    print("\">");
                    printXML(e.lookup(attr));
                    println("</a>");
                }
                indentLevel--;
                print("</c>");
            }
            break;
        }
        case Expr.RELATIVE_TIME:
            // FIXME
            print('\'');
            print(c.value);
            print('\'');
            break;
        case Expr.BOOLEAN:
            print(c == Constant.True
                        ? "<b v=\"t\"/>"
                        : "<b v=\"f\"/>");
            break;
        case Expr.ERROR:
            print("<er/>");
            break;
        case Expr.UNDEFINED:
            print("<un/>");
            break;
        case Expr.INTEGER:
        case Expr.REAL:
            print("<n>");
            print(c.value);
            print("</n>");
            break;
        case Expr.STRING:
            print("<s>");
            printXMLString((String) c.value);
            print("</s>");
            break;
        default:
            print("<e>");
            printNative(exp, true);
            print("</e>");
            break;
        } // switch (exp.type)
    } // printXML(Expr)

    /** Print a string, protecting non-printable and other characters by
     * using escapes.  The output is surrounded by the "quote" character, and
     * inside the string, instances of the quote or backslash as well as
     * non-ASCII or non-printing character are replaced by backslash escapes.
     * If "escapeXML" is true, the four xml magic characters lt, gt, quot, and
     * amp are replaced by extitites such as &amp;lt;.
     * <p>
     * Note: The high-order 8 bits of each character are discarded.
     * @param s the string to be quoted.
     * @param quote an additional charaacter to quote (if non-null).
     * @param escapeXML if true, use xml entities.
     */
    private final void printQuotedString(
                            String s, char quote, boolean escapeXML)
    {
        print(quote);
        for (int i = 0; i < s.length(); i++) {
            int c = s.charAt(i) & 0xff;
            if (c == quote) {
                print("\\");
                print((char) c);
            } else switch (c) {
                case '\n': print("\\n"); break;
                case '\t': print("\\t"); break;
                case '\b': print("\\b"); break;
                case '\r': print("\\r"); break;
                case '\f': print("\\f"); break;
                case '\\': print("\\\\"); break;
                case '<': print(escapeXML ? "&lt;" : "<"); break;
                case '>': print(escapeXML ? "&gt;" : ">"); break;
                case '"': print(escapeXML ? "&quot;" : "\""); break;
                case '&': print(escapeXML ? "&amp;" : "&"); break;
                default:
                    if (c < 040 || c > 0176) {
                        print('\\');
                        print((char) ('0' + ((c >> 6) & 7)));
                        print((char) ('0' + ((c >> 3) & 7)));
                        print((char) ('0' + (c & 7)));
                    } else {
                        print((char) c);
                    }
                    break;
            }
        }
        print(quote);
    } // printQuoted(String)

    /** Print a string, using XML "entities" for non-printing characters and
     * characters with special meaning in XML.
     * <p>
     * Note: The high-order 8 bits of each character are discarded.
     * @param s the string to be quoted.
     * @param quote an additional charaacter to quote (if non-null).
     * @param escapeXML if true, use xml entities.
     */
    private final void printXMLString(String s) {
        for (int i = 0; i < s.length(); i++) {
            int c = s.charAt(i) & 0xff;
            switch (c) {
            case '<': print("&lt;"); break;
            case '>': print("&gt;"); break;
            case '"': print("&quot;"); break;
            case '&': print("&amp;"); break;
            default:
                if (c < 040 || c > 0176) {
                    print("&#");
                    print(c);
                    print(';');
                } else {
                    print((char) c);
                }
                break;
            }
        }
    } // printXMLString(String)
} // ClassAdWriter
