/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2003, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

/**
 * A conditional expression.
 * An internal (non-leaf) node of an expression tree representing the ternary
 * operator <i>cond</i> ? <cond>expr1</cond> : <cond>expr2</cond>.
 * @see Expr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 2.0
 */
public class CondExpr extends Expr {
    private static String VERSION = "$Id: CondExpr.java,v 1.12 2003/01/20 21:05:03 solomon Exp $";

    /** The condition (first component of the expression) */
    public final Expr ec;

    /** The true part (second component of the expression) */
    public final Expr et;

    /** The false part (third component of the expression) */
    public final Expr ef;

    /** Construct a conditional expression from its three components.
     * @param ec the condition (first component of the expression)
     * @param et the true part (second component of the expression)
     * @param ef the false part (third component of the expression)
     */
    public CondExpr(Expr ec, Expr et, Expr ef) {
        super(COND);
        this.ec = ec; this.et = et; this.ef = ef;
    } // CondExpr(Expr,Expr,Expr)

    /** Convert this Expr to a string, appending the result to the end of "sb".
     * The representation is the "canonical native format".
     * @param sb a place to put the result.
     * @return sb.
     * @see ClassAdWriter
     */
    public StringBuffer toString(StringBuffer sb) {
        sb.append('(');
        ec.toString(sb);
        sb.append('?');
        et.toString(sb);
        sb.append(':');
        ef.toString(sb);
        return sb.append(')');
    } // toString()

    /** The precedence of the operator in this expression node.
     * Used to print expressions without superfluous parentheses.
     * @return the precendence of this node.
     * @see ClassAdWriter#MINIMAL_PARENTHESES
     */
    public int prec() {
        return -1;
    } // prec()

    /** Evaluate this Expr.  This is the internal method used to implement
     * {@link Expr#eval()}.
     * @param env an environment used to evaluate the operands.  It is cleared
     * to the null environment before return.
     * @return a Constant representing the value.
     */
    protected Expr eval1(Env env) {
        Env env1 = new Env(env);
        Expr ecv = ec.eval(env1);

        switch (ecv.type) {
        case BOOLEAN:
            return ecv.isTrue() ? et.eval(env) : ef.eval(env);
        case UNDEFINED:
        case ERROR:
            env.clear();
            return ecv;
        default:
            env.clear();
            return Constant.error("type " + ecv.typeName()
                                + " found where boolean expected"
                                + " in conditional expression");
        }
    } // eval1(Env)
} // CondExpr
