/*--------------------------------------------------------------------
 *
 * Condor Java ClassAd library
 * Copyright (C) 1990-2003, CONDOR Team, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and Marvin Solomon
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General
 * Public License as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *--------------------------------------------------------------------*/

package condor.classad;

import java.util.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.text.ParsePosition;


/**
 * A constant (Integer, Real, etc.).  The constant is wrapped so that it
 * may be used as a node in an expression tree.  To conserve space, there
 * is at most one Constant instance with any given value.
 * @see Expr
 * @author <a href="mailto:solomon@cs.wisc.edu">Marvin Solomon</a>
 * @version 2.0
 */
public class Constant extends Expr {
    private static String VERSION = "$Id: Constant.java,v 1.12 2003/01/20 21:05:03 solomon Exp $";

    /** The actual value of this Constant.  Usage depends on type:
     * <dl>
     * <dt>UNDEFINED     <dd>A String error message.
     * <dt>ERROR         <dd>A String error message.
     * <dt>BOOLEAN       <dd>null (there are only two Boolean Constants).
     * <dt>INTEGER       <dd>An Integer value.
     * <dt>REAL          <dd>A Double value.
     * <dt>STRING        <dd>A String value.
     * <dt>ABSOLUTE_TIME <dd>A java.util.Date value.
     * <dt>RELATIVE_TIME <dd>A Long value (number of milliseconds).
     * </dl>
     */
    public final Object value;

    /** The set of all extant instances of Constant, indexed by value.
     * This Map is used to save storage overhead by ensuring that there
     * is at most one Constant with the same value.
     */
    private static Map instanceMap = new HashMap();

    /** Construct an arbitrary constant.
     * Note that this is the only constructor and it is private.
     * Clients create new instances by calling the factory methods
     * newInstance, error, undefined, or bool.
     * @param value the value of this constant.
     * @param type the type of this constant.
     */
    private Constant(Object value, int type) {
        super(type);
        this.value = value;
    } // Constant(Object,int)

    /** A default undefined constant */
    public static final Constant Undef
        = new Constant("", UNDEFINED);

    /** An undefined constant that represents the result a attempting to
     * resovle a cyclic attribute reference as in <code>[a=b;b=a]</code>.
     */
    public static final Constant CyclicRef
                    = undefined("Cyclic attribute definition");

    /** A default error constant */
    public static final Constant Error
        = new Constant("", ERROR);

    /** The unique boolean constant true */
    public static final Constant True
        = new Constant(null, BOOLEAN);

    /** The unique boolean Constant false */
    public static final Constant False
        = new Constant(null, BOOLEAN);

    /** An INTEGER Constant with the value 0, to expedite common cases. */
    static private final Constant INT_ZERO =
        new Constant(new Integer(0), INTEGER);

    /** An INTEGER Constant with the value 1, to expedite common cases. */
    static private final Constant INT_ONE =
        new Constant(new Integer(1), INTEGER);

    /** A REAL Constant with the value 0, to expedite common cases. */
    static private final Constant REAL_ZERO =
        new Constant(new Double(0.0), REAL);

    /** A REAL Constant with the value 1, to expedite common cases. */
    static private final Constant REAL_ONE =
        new Constant(new Double(1.0), REAL);

    /** A STRING Constant with the value "", to expedite common cases. */
    static private final Constant EMPTY_STRING = new Constant("", STRING);

    /** Returns the unique constant of type "type" and value "val", creating
     * one if necessary.  This is the internal method used to implement most of
     * the other getInstance methods
     * @param val the value.
     * @param type the type.
     * @return a new constant.
     */
    private static final Constant getInstance(Object val, int type) {
        Constant result = (Constant) instanceMap.get(val);
        if (result == null) {
            result = new Constant(val, type);
            instanceMap.put(val, result);
        }
        return result;
    } // getInstance(Object, int)

    /** Returns the unique integer constant with value "value", creating one if
     * necessary.
     * @param value the value of the Constant.
     * @return a new INTEGER constant.
     */
    public static final Constant getInstance(int value) {
        if (value == 0) {
            return INT_ZERO;
        }
        if (value == 1) {
            return INT_ONE;
        }
        return getInstance(new Integer(value), INTEGER);
    } // getInstance(int)

    /** Returns the unique real constant with value "value", creating one if
     * necessary.
     * @param value the value of the Constant.
     * @return a new REAL constant.
     */
    public static final Constant getInstance(double value) {
        if (value == 0.0) {
            return REAL_ZERO;
        }
        if (value == 1.0) {
            return REAL_ONE;
        }
        return getInstance(new Double(value), REAL);
    } // getInstance(double)

    /** Returns the unique String constant with value "value", creating one if
     * necessary.
     * @param value the value of the Constant, not including any surrounding
     *              quotes.
     * @return a new STRING constant.
     */
    public static final Constant getInstance(String value) {
        if (value == null) {
            return null;
        }
        if (value.length() == 0) {
            return EMPTY_STRING;
        }
        return getInstance(value, STRING);
    } // getInstance(String)

    /** Create a String Constant from a character array, processing backslash
     * escapes.  Any surrounding quotes have already been stripped.
     * @param buf the array of characters.
     * @param start the index preceding the first character to be used.
     * @param stop the index following the last characater to be used.
     * @return a new STRING constant.
     */
    public static final Constant getInstance(char[] buf, int start, int stop) {
        return getInstance(unquoteString(buf, start, stop));
    } // getInstance(char[],int,int)

    /** Create an absolute time Constant.
     * @param value the value of the Constant.
     * @return a new ABSTIME constant.
     */
    public static final Constant getInstance(Date value) {
        return getInstance(value, ABSOLUTE_TIME);
    } // getInstance(Date)

    /** Returns the unique time constant with value "value", creating one if
     * necessary.
     * @param value the value of the constant in milliseconds
     * @return a new RELTIME constant.
     */
    public static final Constant getInstance(long value) {
        return getInstance(new Long(value), RELATIVE_TIME);
    } // getInstance(long)


    /** Evaluate this Expr.  This is the internal method used to implement
     * {@link Expr#eval()}.
     * @param env an environment used to evaluate the operands.  It is cleared
     * to the null environment before return.
     * @return this Constant.
     */
    protected Expr eval1(Env env) {
        env.clear();
        return this;
    } // eval1(Env)

    /** Convert this Expr to a string, appending the result to the end of "sb".
     * The representation is the "canonical native format".
     * @param sb a place to put the result.
     * @return sb.
     * @see ClassAdWriter
     */
    public StringBuffer toString(StringBuffer sb) {
        switch (type) {
            case UNDEFINED:
                return sb.append("UNDEFINED");
            case ERROR:
                return sb.append("ERROR");
            case BOOLEAN:
                return sb.append(this == True ? "true" : "false");
            case INTEGER:
            case REAL:
                return sb.append(value);
            case STRING:
                return escapeString(sb, (String) value, '"');
            case ABSOLUTE_TIME:
                return sb.append("absTime(\"")
                    .append(dateFormat.format((Date) value))
                    .append("\")");
            case RELATIVE_TIME:
            {
                long v = ((Long) value).longValue();
                return sb.append("relTime(\"")
                    .append(rel2str(v))
                    .append("\"");
            }
            default: throw new RuntimeException("unknown type");
        }
    } // toString(StringBuffer)

    /** Convenience function:  Create an ERROR constant containing a particular
     * message.
     * @param msg the message.
     * @return the Constant.
     */
    public static Constant error(String msg) {
        return new Constant(msg, ERROR);
    } // error(String)

    /** Convenience function:  Create an UNDEFINED constant containing a
     * particular message.
     * @param msg the message.
     * @return the Constant.
     */
    public static Constant undefined(String msg) {
        return new Constant(msg, UNDEFINED);
    } // undefined(String)

    /** Convenience function:  Convert a boolean to Constant.True or
     * Constant.False.
     * @param b the boolean value.
     * @return the Constant.
     */
    public static Constant bool(boolean b) {
        return b ? True : False; 
    } // bool(boolean)

    // Accessors.  Each method coverts the value to the indicated type if
    // possible.  Otherwise, it throws ArithmeticException.

    /** Get the integer value of this Constant if possible.
     * @exception ArithmeticException if this Constant is not an integer.
     * @return the integer value of this IntegerConstant.
     */
    public int intValue() throws ArithmeticException {
        if (type != INTEGER) {
            throw new ArithmeticException(typeName() + " " + this
                + " in integer context");
        }
        return ((Integer)value).intValue();
    } // intValue()

    /** Get the double floating point value of this Constant if possible.
     * @exception ArithmeticException if this Constant is not an instance of
     * IntegerConstant or RealConstant.
     * @return the real value of this RealConstant or the value of this
     * IntegerConstant converted to double.
     */
    public double realValue() throws ArithmeticException {
        switch (type) {
            case REAL:
                return ((Double)value).doubleValue();
            case INTEGER:
                return ((Integer)value).doubleValue();
        }
        throw new ArithmeticException(typeName() + " " + this
            + " in real context");
    } // realValue()

    /** Get the string value of this Constant if possible.
     * @exception ArithmeticException if this Constant is not an instance of
     * StringConstant.
     * @return the string value of this StringConstant.
     */
    public String stringValue() throws ArithmeticException {
        if (type != STRING) {
            throw new ArithmeticException(typeName() + " " + this
                + " in string context");
        }
        return (String)value;
    } // stringValue()

    /** Convenience function to test whether an expression is the constant
     * TRUE.  Note that unlike inValue, etc., this method never throws an
     * exception.
     * @return true if this Expr is a constant of type boolean with value true,
     * and false in all other cases.
     */
    public boolean isTrue() {
        return this == True;
    } // isTrue()

    /** Convert a time value to milliseconds.
     * @exception ArithmeticException if this Constant is not a time value.
     * @return the number of milliseconds (since 1/1/70 for ABSOLUTE_TIME
     * values).
     */
    public long milliseconds() throws ArithmeticException {
        if (type == ABSOLUTE_TIME) {
            return ((Date)value).getTime();
        }
        if (type == RELATIVE_TIME ) {
            return ((Long)value).longValue();
        }
        throw new ArithmeticException(typeName() + " " + this
            + " in integer context");
    } // milliseconds()

    /** The precedence of this expression node (MAXPREC).
     * Used to print expressions without superfluous parentheses.
     * @return the precendence of this node.
     * @see ClassAdWriter#MINIMAL_PARENTHESES
     */
    public int prec() { return MAXPREC+1; }

    /** Generate a string from an array of characters, processing backslash
     * escapes.  The the characters buf[start+1] ... buf[stop-1] are
     * converted to a String, after replacing backslash escapes such as \n by
     * their values.
     * @param buf the array of characters.
     * @param start the index of the first character to be used.
     * @param stop the index following the last characater to be used.
     * @return the converted String.
     */
    /*package*/ static final String unquoteString(char[] buf,
                                        int start, int stop)
    {
        char[] result = new char[stop - start];
        int i = start;
        int j = 0;
        while (i < stop) {
            char c = buf[i++];
            if (c == '\\' && i < stop) {
                c = buf[i++];
                int val = Character.digit(c,8);
                if (val >= 0) {
                    int maxDigits = (val < '4') ? 2 : 1;
                    if (stop - i < maxDigits) {
                        maxDigits = stop - i;
                    }
                    for (int k = 0; k < maxDigits; k++) {
                        c = buf[i];
                        int n = Character.digit(c, 8);
                        if (n < 0) {
                            break;
                        }
                        val = (val << 3) + n;
                        i++;
                    }
                    c = (char) val;
                } else switch (c) {
                    case 'n': c = '\n'; break;
                    case 't': c = '\t'; break;
                    case 'b': c = '\b'; break;
                    case 'r': c = '\r'; break;
                    case 'f': c = '\f'; break;
                    default: break;
                }
            }
            result[j++] = c;
        }
        return new String(result, 0, j);
    } // unquoteString(char[],int,int)

    /** Append to sb a quoted version of string, replacing non-printable values
     * by backslash escapes.  The result starts and ends with quote; any
     * embedded occurrences of quote are preceded by backslashes.
     * Any non-printing characters are escaped with a single-character escape
     * (such as \n) if possible, otherwise \ooo, where ooo is a three-digit
     * octal code.
     * <p>
     * Note that Unicode is not supported: We <em>assume</em> the value of
     * every character is in the range 0..0377.
     * @param sb a place to put the result.
     * @param s the string to be quoted.
     * @param quote the quote character (' or ").
     * @return sb.
     */
    public static StringBuffer
        escapeString(StringBuffer sb, String s, char quote)
    {
        sb.append(quote);
        for (int i=0; i<s.length(); i++) {
            int c = s.charAt(i) & 0xff;
            if (c == quote) {
                sb.append('\\');
            }
            switch (c) {
            case '\n': sb.append('\\').append('n'); break;
            case '\t': sb.append('\\').append('t'); break;
            case '\b': sb.append('\\').append('b'); break;
            case '\r': sb.append('\\').append('r'); break;
            case '\f': sb.append('\\').append('f'); break;
            case '\\': sb.append('\\').append('\\'); break;
            default:
                if (c < 040 || c > 0176) {
                    sb.append('\\');
                    sb.append((char) ('0' + ((c >> 6) & 7)));
                    sb.append((char) ('0' + ((c >> 3) & 7)));
                    sb.append((char) ('0' + (c & 7)));
                } else {
                    sb.append((char) c);
                }
                break;
            }
        }
        return sb.append(quote);
    } // escapeString(String)

    /** Helper table for rel2str. */
    private static int[] threshold = {
        24*60*60*1000,
        60*60*1000,
        60*1000,
        1000
    };

    /** Helper table for rel2str. */
    private static final char[] seperator = { '+', ':', ':' };

    /** Helper table for parseRelTime */
    private static int[] weight = { 0, 24, 60, 60, 1000 };

    /** Helper method for rel2str.
     * Convert n to a decimal string, optionally with leading zeros.
     * @param buf the place to put the result.
     * @param n the number to be converted (must be non-negative).
     * @param m a power of 10 indicating the minimum number of digits in the
     *    result.  For example, if m==100, the result will have at least three
     *    digits.
     */
    private static final void formatDecimal(StringBuffer buf, int n, int m) {
        buf.append((int)(n/m));
        while (m > 1) {
            n %= m;
            m /= 10;
            buf.append((int)(n/m));
        }
    } // formatDecimal(StringBuffer,int,int)

    /** Format a relative time value as a readable String.
     * The result is a string of of the form days+hh:mm:ss.mmm, where
     * leading components are omitted if they are zero.
     * For example,
     * <ul>
     * <li> ten milliseconds becomes ".010".
     * <li> twelve minutes and thirty-four seconds becomes "12:34.000"
     * </ul>
     * @param t the number to be converted, in milliseconds.
     * @return the resulting String.
     */
    /*package*/ static final String rel2str(long t) {
        if (t == 0) {
            return "0";
        }
        StringBuffer buf = new StringBuffer();
        int i = 0;
        if (t < 1000) {
            buf.append("0.");
        } else {
            while (t < threshold[i]) {
                i++;
            }
            formatDecimal(buf, (int)(t/threshold[i]), 1);
            for (;;) {
                t %= threshold[i];
                if (i >= seperator.length) {
                    break;
                }
                buf.append(seperator[i]);
                i++;
                /*
                if (i >= threshold.length) {
                    break;
                }
                */
                formatDecimal(buf, (int)(t/threshold[i]), 10);
            }
        }
        if (t > 0) {
            buf.append('.');
            formatDecimal(buf, (int)t, 100);
        }
        return buf.toString();
    } // rel2str(long)

    /** Pattern for parsing relTime constants. */
    private static Pattern reltimePattern
        = Pattern.compile(
                  "(?:(\\d+)\\+)?"
                + "(?:(?:(\\d+):)?(\\d+):)?"
                + "(\\d+)"
                + "(?:\\.(\\d*))?");

    /** Helper function for relTime(String).  Convert a relative time pattern
     * of the form days+hh:mm:ss[.msec] into a number of milliseconds.
     * @param s the string to parse.
     * @return a number of milliseconds, or -1 for errors.
     */
    /*package*/ static long parseRelTime(String s) {
        try {
            Matcher m = reltimePattern.matcher(s);
            if (!m.matches()) {
                return -1;
            }
            long sum = 0;
            for (int i = 1; i <= weight.length; i++) {
                String f = m.group(i);
                int n = f == null ? 0 : Integer.parseInt(f);
                sum *= weight[i - 1];
                if (i == weight.length && n > 0) {
                    int l = 3 - f.length();
                    if (l < 0) {
                        n = Integer.parseInt(f.substring(0,3));
                    } else {
                        while (l-- > 0) {
                            n *= 10;
                        }
                    }
                }
                sum += n;
            }
            return sum;
        } catch (NumberFormatException e) {
            return -1;
        }
    } // parseRelTime(String)

    /** Cursor for parsing dates and times. */
    private static ParsePosition parsePos = new ParsePosition(0);

    /** Helper function for absTime(String).  Convert an absolute time pattern
     * of the form yyyymmddThhmmss to a Date.
     * @param s the string to parse.
     * @return the corresponding Date, or null for errors.
     */
    /*package*/ static Date parseAbsTime(String s) {
        parsePos.setIndex(0);
        return Expr.dateAndTimeFormat.parse(s, parsePos);
    } // parseAbsTime(String)
} // Constant
